

from __future__ import absolute_import, division, print_function

import pytest
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Circle
from matplotlib.artist import Artist
from numpy.testing import assert_allclose
from matplotlib.backends.backend_agg import FigureCanvasAgg

from glue.tests.helpers import requires_scipy
from glue.utils.misc import DeferredMethod

from ..matplotlib import (point_contour, fast_limits, all_artists, new_artists,
                          remove_artists, view_cascade, get_extent, color2rgb,
                          defer_draw, freeze_margins)


@requires_scipy
class TestPointContour(object):

    def test(self):
        data = np.array([[0, 0, 0, 0],
                         [0, 2, 3, 0],
                         [0, 4, 2, 0],
                         [0, 0, 0, 0]])
        xy = point_contour(2, 2, data)
        x = np.array([2., 2. + 1. / 3., 2., 2., 1, .5, 1, 1, 2])
        y = np.array([2. / 3., 1., 2., 2., 2.5, 2., 1., 1., 2. / 3])

        np.testing.assert_array_almost_equal(xy[:, 0], x)
        np.testing.assert_array_almost_equal(xy[:, 1], y)


def test_fast_limits_nans():
    x = np.zeros((10, 10)) * np.nan
    assert_allclose(fast_limits(x, 0, 1), [0, 1])


def test_single_value():
    x = np.array([1])
    assert_allclose(fast_limits(x, 5., 95.), [1, 1])


def test_artist_functions():

    c1 = Circle((0, 0), radius=1)
    c2 = Circle((1, 0), radius=1)
    c3 = Circle((2, 0), radius=1)

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)

    ax.add_patch(c1)
    ax.add_patch(c2)
    assert all_artists(fig) == set([c1, c2])

    ax.add_patch(c3)
    assert new_artists(fig, set([c1, c2])) == set([c3])

    remove_artists([c2])
    assert all_artists(fig) == set([c1, c3])

    # check that it can deal with being passed the same artist twice
    remove_artists([c1, c1])
    assert all_artists(fig) == set([c3])


def test_get_extent():
    assert get_extent((slice(0, 5, 1), slice(0, 10, 2))) == (0, 10, 0, 5)
    assert get_extent((slice(0, 5, 1), slice(0, 10, 2)), transpose=True) == (0, 5, 0, 10)


def test_view_cascade():
    data = np.zeros((100, 100))

    v2, view = view_cascade(data, (slice(0, 5, 1), slice(0, 5, 1)))
    assert v2 == ((slice(0, 100, 20), slice(0, 100, 20)))
    assert view == (slice(0, 5, 1), slice(0, 5, 1))

    v2, view = view_cascade(data, (3, slice(0, 5, 1)))
    assert v2 == ((3, slice(0, 100, 20)))
    assert view == (3, slice(0, 5, 1))


def test_defer_draw():

    @defer_draw
    def draw_figure():
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.plot([1, 2, 3], [4, 5, 6])
        fig.canvas.draw()
        return 3.5

    result = draw_figure()

    # Make sure that the return value was passed through correctly
    assert result == 3.5


def test_defer_draw_exception():

    # Regression test for a bug that meant that if an exception happened during
    # drawing, the draw method was not restored correctly

    # Make sure we start off with a clean draw method
    assert not isinstance(FigureCanvasAgg.draw, DeferredMethod)

    class ProblematicArtist(Artist):
        def draw(self, *args, **kwargs):
            raise ValueError('You shall not pass!')

    @defer_draw
    def draw_figure():
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.add_artist(ProblematicArtist())
        fig.canvas.draw()

    with pytest.raises(ValueError) as exc:
        result = draw_figure()
    assert exc.value.args[0] == 'You shall not pass!'

    # Make sure that draw is no longer a deferred method
    assert not isinstance(FigureCanvasAgg.draw, DeferredMethod)


@pytest.mark.parametrize(('color', 'rgb'),
                         (('red', (1, 0, 0)), ('green', (0, 0.5020, 0)), ('orange', (1., 0.6470, 0.))))
def test_color2rgb(color, rgb):
    assert_allclose(color2rgb(color), rgb, atol=0.001)


def test_freeze_margins():

    fig = plt.figure(figsize=(4, 4))

    ax = fig.add_subplot(1, 1, 1)
    freeze_margins(ax, margins=[1, 1, 1, 1])

    # Note, we don't test the following since the defaults change depending
    # on the Matplotlib version

    # bbox = ax.get_position()
    # np.testing.assert_allclose(bbox.x0, 0.125)
    # np.testing.assert_allclose(bbox.y0, 0.1)
    # np.testing.assert_allclose(bbox.x1, 0.9)
    # np.testing.assert_allclose(bbox.y1, 0.9)

    fig.canvas.resize_event()

    bbox = ax.get_position()
    np.testing.assert_allclose(bbox.x0, 0.25)
    np.testing.assert_allclose(bbox.y0, 0.25)
    np.testing.assert_allclose(bbox.x1, 0.75)
    np.testing.assert_allclose(bbox.y1, 0.75)

    fig.set_size_inches(8, 8)
    fig.canvas.resize_event()

    bbox = ax.get_position()
    np.testing.assert_allclose(bbox.x0, 0.125)
    np.testing.assert_allclose(bbox.y0, 0.125)
    np.testing.assert_allclose(bbox.x1, 0.875)
    np.testing.assert_allclose(bbox.y1, 0.875)

    ax.resizer.margins = [0, 1, 2, 4]
    fig.canvas.resize_event()

    bbox = ax.get_position()
    np.testing.assert_allclose(bbox.x0, 0.)
    np.testing.assert_allclose(bbox.y0, 0.25)
    np.testing.assert_allclose(bbox.x1, 0.875)
    np.testing.assert_allclose(bbox.y1, 0.5)



"""
Classes to perform aggregations over cubes
"""

from __future__ import absolute_import, division, print_function

try:
    from itertools import izip
except ImportError:  # python3
    izip = zip

from functools import wraps

import numpy as np

from glue.external.six.moves import range as xrange


def check_empty(func):

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.empty_slice:
            return np.zeros(self.shape) * np.nan
        return func(self, *args, **kwargs)

    return wrapper


class Aggregate(object):

    """
    Collapse >=3D datasets into 2D images, using different
    aggregation methods
    """

    def __init__(self, data, attribute, zax, slc, zlim):
        """
        Parameters
        ----------
        data : :class:`~glue.core.data.Data`
        attribute : :class:`~glue.core.component_id.ComponentID`
        zax : int
            Which axis to collapse over.
        slc : tuple of int
            Describes the
            current 2D slice through the image. Used to
            define the orientation, as well as axis values
            for remaining dimensions of >3D cubes
        zlim : tuple of float
            Float values of [lo, hi), describing the limits
            of the slab to collapse over
        """
        self.data = data
        self.attribute = attribute
        self.zax = zax
        self.slc = slc
        self.zlim = min(zlim), max(zlim)

    @property
    def shape(self):
        """
        The shape of the 2D aggregated array
        """
        s = self.data.shape
        return s[self.slc.index('y')], s[self.slc.index('x')]

    @property
    def empty_slice(self):
        """
        True if the slice is empty
        """
        return self.zlim[0] == self.zlim[1]

    def _subslice(self):
        view = [slice(None, None) for _ in self.data.shape]
        ax_collapse = self.zax
        for i, s in enumerate(self.slc):
            if s not in ['x', 'y'] and i != self.zax:
                view[i] = s
                if i < self.zax:
                    ax_collapse -= 1
        view[self.zax] = slice(*self.zlim)
        return view, ax_collapse

    def _prepare_cube(self, attribute=None):
        view, ax_collapse = self._subslice()
        att = attribute or self.attribute
        cube = self.data[att, view]
        return cube, ax_collapse

    def _iter_slice(self, attribute=None):
        # iterate through the uncollapsed slab one plane at a time
        view, ax_collapse = self._subslice()
        att = attribute or self.attribute

        for z in xrange(*self.zlim):
            view[self.zax] = z
            plane = self.data[att, view]
            yield np.nan_to_num(plane)

    def _iter_slice_index(self):
        """Loop over slices of the target attribute and its world coordinate"""
        att = self.data.get_world_component_id(self.zax)
        loop = izip(self._iter_slice(), self._iter_slice(att))
        return loop

    def _finalize(self, cube):
        if self.slc.index('x') < self.slc.index('y'):
            cube = cube.T
        return cube

    def collapse_using(self, function):
        """
        Produce a collapsed image using a numpy aggregation function
        """
        cube, ax = self._prepare_cube()
        result = function(cube, axis=ax)
        return self._finalize(result)

    def _to_world(self, idx):
        args = [None] * self.data.ndim
        y, x = np.mgrid[:idx.shape[0], :idx.shape[1]]
        for i, s in enumerate(self.slc):
            if s not in ['x', 'y']:
                args[i] = np.ones(idx.size) * s
        args[self.slc.index('y')] = y.ravel()
        args[self.slc.index('x')] = x.ravel()
        args[self.zax] = idx.ravel()
        att = self.data.get_world_component_id(self.zax)
        return self.data[att, args].reshape(idx.shape)

    @staticmethod
    def all_operators():
        return (Aggregate.sum,
                Aggregate.mean,
                Aggregate.max,
                Aggregate.argmax,
                Aggregate.argmin,
                Aggregate.mom1,
                Aggregate.mom2,
                Aggregate.median)

    @staticmethod
    def _mean(cube, axis):
        s = np.nansum(cube, axis)
        ct = np.isfinite(cube).sum(axis)
        return 1. * s / ct

    @check_empty
    def sum(self):
        return self.collapse_using(np.nansum)

    @check_empty
    def mean(self):
        return self.collapse_using(self._mean)

    @check_empty
    def max(self):
        return self.collapse_using(np.nanmax)

    @check_empty
    def median(self):
        # NOTE: nans are treated as infinity in this case
        return self.collapse_using(np.median)

    @check_empty
    def argmax(self):
        """
        Location of peak value, in world coords
        """
        idx = self.collapse_using(np.nanargmax)
        return self._to_world(idx)

    @check_empty
    def argmin(self):
        """
        Location of minimum value, in world coords
        """
        idx = self.collapse_using(np.nanargmin)
        return self._to_world(idx)

    @check_empty
    def mom1(self):
        """
        Intensity-weighted coordinate. Pixel units.
        """
        # build up slice-by-slice, to avoid big temporary cubes
        loop = self._iter_slice_index()
        val, loc = next(loop)

        val = np.maximum(val, 0)
        w, result = val, loc * val
        for val, loc in loop:
            val = np.maximum(val, 0)
            result += val * loc
            w += val
        return self._finalize(result / w)

    @check_empty
    def mom2(self):
        """
        Intensity-weighted coordinate dispersion. Pixel units.
        """
        loop = self._iter_slice_index()
        val, loc = next(loop)

        val = np.maximum(val, 0)
        w, x, x2 = val, val * loc, val * loc * loc
        for val, loc in loop:
            val = np.maximum(val, 0)
            w += val
            x += loc * val
            x2 += loc ** 2 * val

        return self._finalize(np.sqrt(x2 / w - (x / w) ** 2))


def mom1(data, axis=0):
    """
    Intensity-weighted coordinate (function version). Pixel units.
    """

    shp = list(data.shape)
    n = shp.pop(axis)

    result = np.zeros(shp)
    w = np.zeros(shp)

    # build up slice-by-slice, to avoid big temporary cubes
    for loc in range(n):
        slc = tuple(loc if j == axis else slice(None) for j in range(data.ndim))
        val = data[slc]
        val = np.maximum(val, 0)
        result += val * loc
        w += val
    return result / w


def mom2(data, axis=0):
    """
    Intensity-weighted coordinate dispersion (function version). Pixel units.
    """

    shp = list(data.shape)
    n = shp.pop(axis)

    x = np.zeros(shp)
    x2 = np.zeros(shp)
    w = np.zeros(shp)

    # build up slice-by-slice, to avoid big temporary cubes
    for loc in range(n):
        slc = tuple(loc if j == axis else slice(None) for j in range(data.ndim))
        val = data[slc]
        val = np.maximum(val, 0)
        x += val * loc
        x2 += val * loc * loc
        w += val
    return np.sqrt(x2 / w - (x / w) ** 2)



# pylint: disable=I0011,W0613,W0201,W0212,E1101,E1103

from __future__ import absolute_import, division, print_function

import os

import pytest
import numpy as np

from numpy.testing import assert_equal

from glue.core import Data
from glue.core.roi import PointROI
from glue.viewers.matplotlib.qt.tests.test_data_viewer import BaseTestMatplotlibDataViewer
from glue.app.qt import GlueApplication
from glue.core.state import GlueUnSerializer

from ..data_viewer import DendrogramViewer

DATA = os.path.join(os.path.dirname(__file__), 'data')


class TestDendrogramCommon(BaseTestMatplotlibDataViewer):

    viewer_cls = DendrogramViewer

    def init_data(self):
        return Data(label='d1', parent=[-1, 0, 1, 1], height=[1.3, 2.2, 3.2, 4.4])

    # TODO: Find a way to simplify having to overload this just because
    # we need to use a different ROI
    def test_apply_roi_undo(self):
        pass

        self.data_collection.append(self.data)
        self.viewer.add_data(self.data)

        roi = PointROI(0.2, 2.8)
        self.viewer.apply_roi(roi)

        assert len(self.data.subsets) == 1

        mask1 = self.data.subsets[0].subset_state.to_mask(self.data)

        roi = PointROI(0.7, 2.8)
        self.viewer.apply_roi(roi)

        assert len(self.data.subsets) == 1

        mask2 = self.data.subsets[0].subset_state.to_mask(self.data)

        assert np.any(mask1 != mask2)

        self.application.undo()

        assert len(self.data.subsets) == 1

        mask3 = self.data.subsets[0].subset_state.to_mask(self.data)

        assert np.all(mask3 == mask1)

        self.application.redo()

        assert len(self.data.subsets) == 1

        mask4 = self.data.subsets[0].subset_state.to_mask(self.data)

        assert np.all(mask4 == mask2)

    @pytest.mark.skip
    def test_add_invalid_data():
        pass


class TestDendrogramViewer():

    def setup_method(self, method):

        self.data = Data(label='d1', parent=[-1, 0, 1, 1], height=[1.3, 2.2, 3.2, 4.4])

        self.app = GlueApplication()
        self.session = self.app.session
        self.hub = self.session.hub

        self.data_collection = self.session.data_collection
        self.data_collection.append(self.data)

        self.viewer = self.app.new_data_viewer(DendrogramViewer)

        self.data_collection.register_to_hub(self.hub)
        self.viewer.register_to_hub(self.hub)

    def teardown_method(self, method):
        self.viewer.close()

    def test_point_select(self):

        self.viewer.add_data(self.data)

        # By default selecting a structure selects all substructures
        roi = PointROI(0.5, 1.5)
        self.viewer.apply_roi(roi)
        assert len(self.data.subsets) == 1
        mask1 = self.data.subsets[0].subset_state.to_mask(self.data)
        assert_equal(mask1, [0, 1, 1, 1])

        # But this option can be turned off
        self.viewer.state.select_substruct = False
        self.viewer.apply_roi(roi)
        assert len(self.data.subsets) == 1
        mask1 = self.data.subsets[0].subset_state.to_mask(self.data)
        assert_equal(mask1, [0, 1, 0, 0])
        self.viewer.state.select_substruct = True

        # Try selecting a leaf
        roi = PointROI(0.2, 2.8)
        self.viewer.apply_roi(roi)
        assert len(self.data.subsets) == 1
        mask1 = self.data.subsets[0].subset_state.to_mask(self.data)
        assert_equal(mask1, [0, 0, 1, 0])

        # Try selecting another leaf
        roi = PointROI(0.7, 2.8)
        self.viewer.apply_roi(roi)
        assert len(self.data.subsets) == 1
        mask1 = self.data.subsets[0].subset_state.to_mask(self.data)
        assert_equal(mask1, [0, 0, 0, 1])

    def test_attribute_change_triggers_relayout(self):

        self.data.add_component([4, 5, 6, 7], 'flux')
        self.viewer.add_data(self.data)

        l = self.viewer.state._layout
        self.viewer.state.height_att = self.data.id['flux']
        assert self.viewer.state._layout is not l


class TestSessions(object):

    @pytest.mark.parametrize('protocol', [0, 1])
    def test_session_back_compat(self, protocol):

        filename = os.path.join(DATA, 'dendro_v{0}.glu'.format(protocol))

        with open(filename, 'r') as f:
            session = f.read()

        state = GlueUnSerializer.loads(session)

        ga = state.object('__main__')

        dc = ga.session.data_collection

        assert len(dc) == 1

        assert dc[0].label == 'data'

        viewer1 = ga.viewers[0][0]

        assert len(viewer1.state.layers) == 2

        assert viewer1.state.parent_att is dc[0].id['parent']
        assert viewer1.state.height_att is dc[0].id['height']
        assert viewer1.state.order_att is dc[0].id['height']

        layer_state = viewer1.state.layers[0]
        assert layer_state.visible
        assert layer_state.layer is dc[0]

        layer_state = viewer1.state.layers[1]
        assert layer_state.visible
        assert layer_state.layer is dc[0].subsets[0]

        ga.close()



from __future__ import absolute_import, division, print_function

from itertools import chain
from weakref import WeakKeyDictionary

from .core import CallbackProperty

__all__ = ['ChoiceSeparator', 'SelectionCallbackProperty']


class ChoiceSeparator(str):
    pass


class SelectionCallbackProperty(CallbackProperty):

    def __init__(self, default_index=0, **kwargs):
        super(SelectionCallbackProperty, self).__init__(**kwargs)
        self.default_index = default_index
        self._choices = WeakKeyDictionary()
        self._display = WeakKeyDictionary()

    def __set__(self, instance, value):
        if value is not None and value not in self._choices.get(instance, ()):
            raise ValueError('value {0} is not in valid choices'.format(value))
        super(SelectionCallbackProperty, self).__set__(instance, value)

    def _get_full_info(self, instance):
        return self.__get__(instance), self._choices.get(instance, ())

    def get_display_func(self, instance):
        return self._display.get(instance, None)

    def set_display_func(self, instance, display):
        self._display[instance] = display
        # selection = self.__get__(instance)
        # self.notify(instance, selection, selection)

    def get_choices(self, instance):
        return self._choices.get(instance, ())

    def get_choice_labels(self, instance):
        display = self._display.get(instance, str)
        labels = []
        for choice in self.get_choices(instance):
            if isinstance(choice, ChoiceSeparator):
                labels.append(str(choice))
            else:
                labels.append(display(choice))
        return labels

    def set_choices(self, instance, choices):
        self._choices[instance] = choices
        self._choices_updated(instance, choices)
        selection = self.__get__(instance)
        self.notify(instance, selection, selection)

    def _choices_updated(self, instance, choices):

        if not choices:
            self.__set__(instance, None)
            return

        selection = self.__get__(instance)

        if selection in choices:
            return

        if self.default_index < 0:
            index = len(choices) + self.default_index
        else:
            index = self.default_index

        index = min(index, len(choices) - 1)

        if isinstance(choices[index], ChoiceSeparator):
            for i in chain(range(index + 1, len(choices)), range(index - 1, -1, -1)):
                if not isinstance(choices[i], ChoiceSeparator):
                    index = i
                    break
            else:
                self.__set__(instance, None)
                return

        self.__set__(instance, choices[index])



from __future__ import absolute_import, division, print_function

import traceback

__all__ = ['die_on_error', 'avoid_circular']


def die_on_error(msg):
    """
    Non-GUI version of the decorator in glue.utils.qt.decorators.

    In this case we just let the Python exception terminate the execution.
    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                print('=' * 72)
                print(msg + ' (traceback below)')
                print('-' * 72)
                traceback.print_exc()
                print('=' * 72)
        return wrapper
    return decorator


def avoid_circular(meth):
    def wrapper(self, *args, **kwargs):
        if not hasattr(self, '_in_avoid_circular') or not self._in_avoid_circular:
            self._in_avoid_circular = True
            try:
                return meth(self, *args, **kwargs)
            finally:
                self._in_avoid_circular = False
    return wrapper



from __future__ import absolute_import, division, print_function

import os
import logging
import platform
import traceback

import numpy as np

from qtpy import QtCore, QtGui, QtWidgets, compat
from qtpy.QtCore import Qt

from glue.external.six.moves import range as xrange
from glue.core.exceptions import IncompatibleAttribute
from glue.core import Subset
from glue.core.callback_property import add_callback, ignore_callback
from glue.config import fit_plugin, viewer_tool
from glue.viewers.matplotlib.qt.toolbar import MatplotlibViewerToolbar
from glue.core.qt.mime import LAYERS_MIME_TYPE
from glue.viewers.common.qt.mouse_mode import RoiMode
from glue.utils.qt import load_ui, get_qapp
from glue.core.qt.simpleforms import build_form_item
from glue.utils.qt.widget_properties import CurrentComboProperty
from glue.app.qt.mdi_area import GlueMdiSubWindow
from glue.viewers.matplotlib.qt.widget import MplWidget
from glue.utils import nonpartial, Pointer
from glue.utils.qt import Worker, messagebox_on_error
from glue.core.subset import RoiSubsetState
from glue.core.qt import roi as qt_roi
from .profile_viewer import ProfileViewer
from glue.viewers.image.state import AggregateSlice
from glue.core.aggregate import mom1, mom2


class Extractor(object):
    # Warning:
    # Coordinate conversion is not well-defined if pix2world is not
    # monotonic!

    @staticmethod
    def abcissa(data, axis):
        slc = [0 for _ in data.shape]
        slc[axis] = slice(None, None)
        att = data.get_world_component_id(axis)
        return data[att, tuple(slc)].ravel()

    @staticmethod
    def spectrum(data, attribute, roi, slc, zaxis):

        # Find the integer index of the x and y axes, which are the axes for
        # which the image is shown (the ROI is drawn along these attributes)
        xaxis = slc.index('x')
        yaxis = slc.index('y')

        # Get the actual component IDs corresponding to these axes
        xatt = data.get_pixel_component_id(xaxis)
        yatt = data.get_pixel_component_id(yaxis)

        # Set up a view that does not reduce the dimensionality of the array but
        # extracts 1-element slices along dimensions that are not relevant.
        view = []
        for idim, dim in enumerate(slc):
            if idim in (xaxis, yaxis, zaxis):
                view.append(slice(None))
            else:
                view.append(slice(dim, dim + 1))
        view = tuple(view)

        # We now delegate to RoiSubsetState to compute the mask based on the ROI
        subset_state = RoiSubsetState(xatt=xatt, yatt=yatt, roi=roi)
        mask = subset_state.to_mask(data, view=view)

        # We now extract the values that fall inside the ROI. Unfortunately,
        # this returns a flat 1-d array, so we need to then reshape it to get
        # an array with shape (n_spec, n_pix), where n_pix is the number of
        # pixels inside the ROI

        values = data[attribute, view]

        if zaxis != 0:
            values = values.swapaxes(zaxis, 0)
            mask = mask.swapaxes(zaxis, 0)

        values = values[mask].reshape(data.shape[zaxis], -1)

        # We then average along the spatial dimension
        spectrum = np.nanmean(values, axis=1)

        # Get the world coordinates of the spectral axis
        x = Extractor.abcissa(data, zaxis)

        return x, spectrum

    @staticmethod
    def world2pixel(data, axis, value):
        x = Extractor.abcissa(data, axis)
        if x.size > 1 and (x[1] < x[0]):
            x = x[::-1]
            result = x.size - np.searchsorted(x, value) - 2
        else:
            result = np.searchsorted(x, value) - 1
        return np.clip(result, 0, x.size - 1)

    @staticmethod
    def pixel2world(data, axis, value):
        x = Extractor.abcissa(data, axis)
        return x[np.clip(value, 0, x.size - 1)]

    @staticmethod
    def subset_spectrum(subset, attribute, slc, zaxis):
        """
        Extract a spectrum from a subset.

        This makes a mask of the subset in the **current slice**,
        and extracts a tube of this shape over all slices along ``zaxis``.
        In other words, the variation of the subset along ``zaxis`` is ignored,
        and only the interaction of the subset and the slice is relevant.

        :param subset: A :class:`~glue.core.subset.Subset`
        :param attribute: The :class:`~glue.core.data.ComponentID` to extract
        :param slc: A tuple describing the slice
        :param zaxis: Which axis to integrate over
        """
        data = subset.data
        x = Extractor.abcissa(data, zaxis)

        view = [slice(s, s + 1)
                if s not in ['x', 'y'] else slice(None)
                for s in slc]

        mask = np.squeeze(subset.to_mask(view))
        if slc.index('x') < slc.index('y'):
            mask = mask.T

        w = np.where(mask)
        view[slc.index('x')] = w[1]
        view[slc.index('y')] = w[0]

        result = np.empty(x.size)

        # treat each channel separately, to reduce memory storage
        for i in xrange(data.shape[zaxis]):
            view[zaxis] = i
            val = data[attribute, view]
            result[i] = np.nansum(val) / np.isfinite(val).sum()

        y = result

        return x, y


class SpectrumContext(object):

    """
    Base class for different interaction contexts
    """
    viewer_state = Pointer('main.viewer_state')
    data = Pointer('main.data')
    profile_axis = Pointer('main.profile_axis')
    canvas = Pointer('main.canvas')
    profile = Pointer('main.profile')

    def __init__(self, main):

        self.main = main
        self.grip = None
        self.panel = None
        self.widget = None

        self._setup_grip()
        self._setup_widget()
        self._connect()

    def _setup_grip(self):
        """ Create a :class:`~glue.plugins.tools.spectrum_tool.profile_viewer.Grip` object
            to interact with the plot. Assign to self.grip
        """
        raise NotImplementedError()

    def _setup_widget(self):
        """
        Create a context-specific widget
        """
        # this is the widget that is displayed to the right of the
        # spectrum
        raise NotImplementedError()

    def _connect(self):
        """
        Attach event handlers
        """
        pass

    def set_enabled(self, enabled):
        self.enable() if enabled else self.disable()

    def enable(self):
        if self.grip is not None:
            self.grip.enable()

    def disable(self):
        if self.grip is not None:
            self.grip.disable()

    def recenter(self, lim):
        """Re-center the grip to the given x axlis limit tuple"""
        if self.grip is None:
            return
        if hasattr(self.grip, 'value'):
            self.grip.value = sum(lim) / 2.
            return

        # Range grip
        cen = sum(lim) / 2
        wid = max(lim) - min(lim)
        self.grip.range = cen - wid / 4, cen + wid / 4


class NavContext(SpectrumContext):

    """
    Mode to set the 2D slice in the parent image widget by dragging
    a handle in the spectrum
    """

    def _setup_grip(self):
        def _set_state_from_grip(value):
            """Update state.slices given grip value"""
            if not self.main.enabled:
                return

            slc = list(self.viewer_state.slices)
            # state.slices stored in pixel coords
            value = Extractor.world2pixel(
                self.data,
                self.profile_axis, value)
            slc[self.profile_axis] = value

            # prevent callback bouncing. Fixes #298
            self.viewer_state.slices = tuple(slc)

        def _set_grip_from_state(slc):
            """Update grip.value given state.slices"""
            if not self.main.enabled:
                return

            # grip.value is stored in world coordinates
            val = slc[self.profile_axis]

            if isinstance(val, AggregateSlice):
                val = val.center

            val = Extractor.pixel2world(self.data, self.profile_axis, val)

            # If pix2world not monotonic, this can trigger infinite recursion.
            # Avoid by disabling callback loop
            # XXX better to specifically ignore _set_state_from_grip
            with ignore_callback(self.grip, 'value'):
                self.grip.value = val

        self.grip = self.main.profile.new_value_grip()

        add_callback(self.viewer_state, 'slices', _set_grip_from_state)
        add_callback(self.grip, 'value', _set_state_from_grip)

    def _connect(self):
        pass

    def _setup_widget(self):
        self.widget = QtWidgets.QTextEdit()
        self.widget.setHtml("To <b> slide </b> through the cube, "
                            "drag the handle or double-click<br><br><br>"
                            "To make a <b> new profile </b>, "
                            "click-drag a new box in the image, or drag "
                            "a subset onto the plot to the left")
        self.widget.setTextInteractionFlags(Qt.NoTextInteraction)


class CollapseContext(SpectrumContext):
    """
    Mode to collapse a section of a cube into a 2D image.

    Supports several aggregations: mean, median, max, mom1, mom2
    """

    def _setup_grip(self):
        self.grip = self.main.profile.new_range_grip()

    def _setup_widget(self):
        w = QtWidgets.QWidget()
        l = QtWidgets.QFormLayout()
        w.setLayout(l)

        combo = QtWidgets.QComboBox()
        combo.addItem("Mean", userData=np.mean)
        combo.addItem("Median", userData=np.median)
        combo.addItem("Max", userData=np.max)
        combo.addItem("Centroid", userData=mom1)
        combo.addItem("Linewidth", userData=mom2)

        run = QtWidgets.QPushButton("Collapse")
        save = QtWidgets.QPushButton("Save as FITS file")

        buttons = QtWidgets.QHBoxLayout()
        buttons.addWidget(run)
        buttons.addWidget(save)

        self._save = save
        self._run = run

        l.addRow("", combo)
        l.addRow("", buttons)

        self.widget = w
        self._combo = combo

        self._collapsed_viewer = None

    def _connect(self):
        self._run.clicked.connect(nonpartial(self._aggregate))
        self._save.clicked.connect(nonpartial(self._choose_save))

    @property
    def aggregator(self):
        return self._combo.itemData(self._combo.currentIndex())

    @property
    def aggregator_label(self):
        return self._combo.currentText()

    def _aggregate(self):

        func = self.aggregator

        rng = list(self.grip.range)

        rng = Extractor.world2pixel(self.data,
                                    self.profile_axis,
                                    rng)
        rng[1] += 1

        slices = list(self.viewer_state.slices)

        current_slice = slices[self.profile_axis]
        if isinstance(current_slice, AggregateSlice):
            current_slice = current_slice.center

        slices[self.profile_axis] = AggregateSlice(slice(*rng),
                                                   current_slice,
                                                   func)

        self.viewer_state.slices = tuple(slices)

        # Save a local copy of the collapsed array
        for layer_state in self.viewer_state.layers:
            if layer_state.layer is self.viewer_state.reference_data:
                break
        else:
            raise Exception("Couldn't find layer corresponding to reference data")

        self._agg = layer_state.get_sliced_data()

    @messagebox_on_error("Failed to export projection")
    def _choose_save(self):

        self._aggregate()

        out, _ = compat.getsavefilename(filters='FITS Files (*.fits)')
        if not out:
            return

        self.save_to(out)

    def save_to(self, pth):
        """
        Write the projection to a file

        Parameters
        ----------
        pth : str
           Path to write to
        """

        from astropy.io import fits

        data = self.viewer_state.reference_data
        if data is None:
            raise RuntimeError("Cannot save projection -- no data to visualize")

        self._aggregate()

        # try to project wcs to 2D
        wcs = getattr(data.coords, 'wcs', None)
        if wcs:
            try:
                wcs.dropaxis(data.ndim - 1 - self.main.profile_axis)
                header = wcs.to_header(True)
            except Exception as e:
                msg = "Could not extract 2D wcs for this data: %s" % e
                logging.getLogger(__name__).warn(msg)
                header = fits.Header()
        else:
            header = fits.Header()

        lo, hi = self.grip.range
        history = ('Created by Glue. %s projection over channels %i-%i of axis %i. Slice=%s' %
                   (self.aggregator_label, lo, hi, self.main.profile_axis, self.viewer_state.slices))

        header.add_history(history)

        try:
            fits.writeto(pth, self._agg, header, overwrite=True)
        except TypeError:
            fits.writeto(pth, self._agg, header, clobber=True)


class ConstraintsWidget(QtWidgets.QWidget):

    """
    A widget to display and tweak the constraints of a :class:`~glue.core.fitters.BaseFitter1D`
    """

    def __init__(self, constraints, parent=None):
        """
        Parameters
        ----------
        constraints : dict
            The `contstraints` property of a :class:`~glue.core.fitters.BaseFitter1D`
            object
        parent : QtWidgets.QWidget (optional)
            The parent of this widget
        """
        super(ConstraintsWidget, self).__init__(parent)
        self.constraints = constraints

        self.layout = QtWidgets.QGridLayout()
        self.layout.setContentsMargins(2, 2, 2, 2)
        self.layout.setSpacing(4)

        self.setLayout(self.layout)

        self.layout.addWidget(QtWidgets.QLabel("Estimate"), 0, 1)
        self.layout.addWidget(QtWidgets.QLabel("Fixed"), 0, 2)
        self.layout.addWidget(QtWidgets.QLabel("Bounded"), 0, 3)
        self.layout.addWidget(QtWidgets.QLabel("Lower Bound"), 0, 4)
        self.layout.addWidget(QtWidgets.QLabel("Upper Bound"), 0, 5)

        self._widgets = {}
        names = sorted(list(self.constraints.keys()))

        for k in names:
            row = []
            w = QtWidgets.QLabel(k)
            row.append(w)

            v = QtGui.QDoubleValidator()
            e = QtWidgets.QLineEdit()
            e.setValidator(v)
            e.setText(str(constraints[k]['value'] or ''))
            row.append(e)

            w = QtWidgets.QCheckBox()
            w.setChecked(constraints[k]['fixed'])
            fix = w
            row.append(w)

            w = QtWidgets.QCheckBox()
            limits = constraints[k]['limits']
            w.setChecked(limits is not None)
            bound = w
            row.append(w)

            e = QtWidgets.QLineEdit()
            e.setValidator(v)
            if limits is not None:
                e.setText(str(limits[0]))
            row.append(e)

            e = QtWidgets.QLineEdit()
            e.setValidator(v)
            if limits is not None:
                e.setText(str(limits[1]))
            row.append(e)

            def unset(w):
                def result(active):
                    if active:
                        w.setChecked(False)
                return result

            fix.toggled.connect(unset(bound))
            bound.toggled.connect(unset(fix))

            self._widgets[k] = row

        for i, row in enumerate(names, 1):
            for j, widget in enumerate(self._widgets[row]):
                self.layout.addWidget(widget, i, j)

    def settings(self, name):
        """ Return the constraints for a single model parameter """
        row = self._widgets[name]
        name, value, fixed, limited, lo, hi = row
        value = float(value.text()) if value.text() else None
        fixed = fixed.isChecked()
        limited = limited.isChecked()
        lo = lo.text()
        hi = hi.text()
        limited = limited and not ((not lo) or (not hi))
        limits = None if not limited else [float(lo), float(hi)]
        return dict(value=value, fixed=fixed, limits=limits)

    def update_constraints(self, fitter):
        """ Update the constraints in a :class:`~glue.core.fitters.BaseFitter1D`
            based on the settings in this widget
        """
        for name in self._widgets:
            s = self.settings(name)
            fitter.set_constraint(name, **s)


class FitSettingsWidget(QtWidgets.QDialog):

    def __init__(self, fitter, parent=None):
        super(FitSettingsWidget, self).__init__(parent)
        self.fitter = fitter

        self._build_form()
        self._connect()
        self.setModal(True)

    def _build_form(self):
        fitter = self.fitter

        l = QtWidgets.QFormLayout()
        options = fitter.options
        self.widgets = {}
        self.forms = {}

        for k in sorted(options):
            item = build_form_item(fitter, k)
            l.addRow(item.label, item.widget)
            self.widgets[k] = item.widget
            self.forms[k] = item  # need to prevent garbage collection

        constraints = fitter.constraints
        if constraints:
            self.constraints = ConstraintsWidget(constraints)
            l.addRow(self.constraints)
        else:
            self.constraints = None

        self.okcancel = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok |
                                               QtWidgets.QDialogButtonBox.Cancel)
        l.addRow(self.okcancel)
        self.setLayout(l)

    def _connect(self):
        self.okcancel.accepted.connect(self.accept)
        self.okcancel.rejected.connect(self.reject)
        self.accepted.connect(self.update_fitter_from_settings)

    def update_fitter_from_settings(self):
        for k, v in self.widgets.items():
            setattr(self.fitter, k, v.value())
        if self.constraints is not None:
            self.constraints.update_constraints(self.fitter)


class FitContext(SpectrumContext):

    """
    Mode to fit a range of a spectrum with a model fitter.

    Fitters are taken from user-defined fit plugins, or
    :class:`~glue.core.fitters.BaseFitter1D` subclasses
    """
    error = CurrentComboProperty('ui.uncertainty_combo')
    fitter = CurrentComboProperty('ui.profile_combo')

    def _setup_grip(self):
        self.grip = self.main.profile.new_range_grip()

    def _setup_widget(self):
        self.ui = load_ui('spectrum_fit_panel.ui', None,
                          directory=os.path.dirname(__file__))
        self.ui.uncertainty_combo.hide()
        self.ui.uncertainty_label.hide()
        font = QtGui.QFont("Courier")
        font.setStyleHint(font.Monospace)
        self.ui.results_box.document().setDefaultFont(font)
        self.ui.results_box.setLineWrapMode(self.ui.results_box.NoWrap)
        self.widget = self.ui

        for fitter in list(fit_plugin):
            self.ui.profile_combo.addItem(fitter.label,
                                          userData=fitter())

    def _edit_model_options(self):

        d = FitSettingsWidget(self.fitter)
        d.exec_()

    def _connect(self):
        self.ui.fit_button.clicked.connect(nonpartial(self.fit))
        self.ui.clear_button.clicked.connect(nonpartial(self.clear))
        self.ui.settings_button.clicked.connect(
            nonpartial(self._edit_model_options))

    def fit(self):
        """
        Fit a model to the data

        The fitting happens on a dedicated thread, to keep the UI
        responsive
        """
        xlim = self.grip.range
        fitter = self.fitter

        def on_success(result):
            fit_result, _, _, _ = result
            self._report_fit(fitter.summarize(*result))
            self.main.profile.plot_fit(fitter, fit_result)

        def on_fail(exc_info):
            exc = '\n'.join(traceback.format_exception(*exc_info))
            self._report_fit("Error during fitting:\n%s" % exc)

        def on_done():
            self.ui.fit_button.setText("Fit")
            self.ui.fit_button.setEnabled(True)
            self.canvas.draw()

        self.ui.fit_button.setText("Running...")
        self.ui.fit_button.setEnabled(False)

        w = Worker(self.main.profile.fit, fitter, xlim=xlim)
        w.result.connect(on_success)
        w.error.connect(on_fail)
        w.finished.connect(on_done)

        self._fit_worker = w  # hold onto a reference
        w.start()

    def _report_fit(self, report):
        self.ui.results_box.document().setPlainText(report)

    def clear(self):
        self.ui.results_box.document().setPlainText('')
        self.main.profile.clear_fit()
        self.canvas.draw()


class SpectrumMainWindow(QtWidgets.QMainWindow):

    """
    The main window that the spectrum viewer is embedded in.

    Defines two signals to trigger when a subset is dropped into the window,
    and when the window is closed.
    """
    subset_dropped = QtCore.Signal(object)
    window_closed = QtCore.Signal()

    def __init__(self, parent=None):
        super(SpectrumMainWindow, self).__init__(parent=parent)
        self.setAcceptDrops(True)

    def closeEvent(self, event):
        self.window_closed.emit()
        return super(SpectrumMainWindow, self).closeEvent(event)

    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat(LAYERS_MIME_TYPE):
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        layer = event.mimeData().data(LAYERS_MIME_TYPE)[0]
        if isinstance(layer, Subset):
            self.subset_dropped.emit(layer)

    def set_status(self, message):
        sb = self.statusBar()
        sb.showMessage(message)

@viewer_tool
class SpectrumExtractorMode(RoiMode):

    """
    Lets the user select a region in an image and, when connected to a
    SpectrumExtractorTool, uses this to display spectra extracted from that
    position
    """
    persistent = True

    icon = 'glue_spectrum'
    tool_id = 'spectrum'
    action_text = 'Spectrum'
    tool_tip = 'Extract a spectrum from the selection'
    shortcut = 'S'

    def __init__(self, viewer, **kwargs):
        super(SpectrumExtractorMode, self).__init__(viewer, **kwargs)
        self._roi_tool = qt_roi.QtRectangularROI(self._axes)  # default
        self._tool = SpectrumTool(self.viewer, self)
        self._release_callback = self._tool._update_profile
        self._move_callback = self._tool._move_profile
        self._roi_callback = None
        self.viewer.state.add_callback('reference_data', self._on_reference_data_change)

    def _on_reference_data_change(self, reference_data):
        if reference_data is not None:
            self.enabled = reference_data.ndim == 3

    def menu_actions(self):

        result = []

        a = QtWidgets.QAction('Rectangle', None)
        a.triggered.connect(nonpartial(self.set_roi_tool, 'Rectangle'))
        result.append(a)

        a = QtWidgets.QAction('Circle', None)
        a.triggered.connect(nonpartial(self.set_roi_tool, 'Circle'))
        result.append(a)

        a = QtWidgets.QAction('Polygon', None)
        a.triggered.connect(nonpartial(self.set_roi_tool, 'Polygon'))
        result.append(a)

        for r in result:
            if self._move_callback is not None:
                r.triggered.connect(nonpartial(self._move_callback, self))

        return result

    def set_roi_tool(self, mode):
        if mode is 'Rectangle':
            self._roi_tool = qt_roi.QtRectangularROI(self._axes)

        if mode is 'Circle':
            self._roi_tool = qt_roi.QtCircularROI(self._axes)

        if mode is 'Polygon':
            self._roi_tool = qt_roi.QtPolygonalROI(self._axes)

        self._roi_tool.plot_opts.update(edgecolor='#c51b7d',
                                        facecolor=None,
                                        edgewidth=3,
                                        alpha=1.0)

    def close(self):
        self._tool.close()
        return super(SpectrumExtractorMode, self).close()

# TODO: refactor this so that we don't have a separate tool and mode


class SpectrumTool(object):

    """
    Main widget for interacting with spectra extracted from an image.

    Provides different contexts for interacting with the spectrum:

    *navigation context* lets the user set the slice in the parent image
                         by dragging a bar on the spectrum
    *fit context* lets the user fit models to a portion of the spectrum
    *collapse context* lets the users collapse a section of a cube to a 2D image
    """

    def __init__(self, image_viewer, mouse_mode):
        self._relim_requested = True

        self.image_viewer = image_viewer
        self.viewer_state = self.image_viewer.state
        self.image_viewer.window_closed.connect(self.close)

        self._build_main_widget()

        self.profile = ProfileViewer(self.canvas.fig)
        self.axes = self.profile.axes

        self.mouse_mode = mouse_mode
        self._setup_toolbar()

        self._setup_ctxbar()

        self._connect()
        w = self.image_viewer.session.application.add_widget(self,
                                                             label='Profile')
        w.close()

    def close(self):
        if hasattr(self, '_mdi_wrapper'):
            self._mdi_wrapper.close()
        else:
            self.widget.close()
        self.image_viewer = None

    @property
    def enabled(self):
        """Return whether the window is visible and active"""
        return self.widget.isVisible()

    def mdi_wrap(self):
        sub = GlueMdiSubWindow()
        sub.setWidget(self.widget)
        self.widget.destroyed.connect(sub.close)
        sub.resize(self.widget.size())
        self._mdi_wrapper = sub
        return sub

    def _build_main_widget(self):
        self.widget = SpectrumMainWindow()
        self.widget.window_closed.connect(self.reset)

        w = QtWidgets.QWidget()
        l = QtWidgets.QHBoxLayout()
        l.setSpacing(2)
        l.setContentsMargins(2, 2, 2, 2)
        w.setLayout(l)

        mpl = MplWidget()
        self.canvas = mpl.canvas
        l.addWidget(mpl)
        l.setStretchFactor(mpl, 5)

        self.widget.setCentralWidget(w)

        # TODO: fix hacks
        w.canvas = self.canvas
        self.widget.central_widget = w

    def _setup_ctxbar(self):
        l = self.widget.centralWidget().layout()
        self._contexts = [NavContext(self),
                          FitContext(self),
                          CollapseContext(self)]

        tabs = QtWidgets.QTabWidget(parent=self.widget)

        # The following is needed because of a bug in Qt which means that
        # tab titles don't get scaled right.
        if platform.system() == 'Darwin':
            app = get_qapp()
            app_font = app.font()
            tabs.setStyleSheet('font-size: {0}px'.format(app_font.pointSize()))

        tabs.addTab(self._contexts[0].widget, 'Navigate')
        tabs.addTab(self._contexts[1].widget, 'Fit')
        tabs.addTab(self._contexts[2].widget, 'Collapse')
        self._tabs = tabs
        self._tabs.setVisible(False)
        l.addWidget(tabs)
        l.setStretchFactor(tabs, 0)

    def _connect(self):

        add_callback(self.viewer_state, 'x_att',
                     self.reset)
        add_callback(self.viewer_state, 'y_att',
                     self.reset)

        def _on_tab_change(index):
            for i, ctx in enumerate(self._contexts):
                ctx.set_enabled(i == index)
                if i == index:
                    self.profile.active_grip = ctx.grip

        self._tabs.currentChanged.connect(_on_tab_change)
        _on_tab_change(self._tabs.currentIndex())

        self.widget.subset_dropped.connect(self._extract_subset_profile)

    def _setup_toolbar(self):

        tb = MatplotlibViewerToolbar(self.widget)

        # disable ProfileViewer mouse processing during mouse modes
        tb.tool_activated.connect(self.profile.disconnect)
        tb.tool_deactivated.connect(self.profile.connect)

        self._menu_toggle_action = QtWidgets.QAction("Options", tb)
        self._menu_toggle_action.setCheckable(True)
        self._menu_toggle_action.toggled.connect(self._toggle_menu)

        tb.addAction(self._menu_toggle_action)
        self.widget.addToolBar(tb)
        return tb

    def _toggle_menu(self, active):
        self._tabs.setVisible(active)

    def reset(self, *args):
        self.hide()
        self.mouse_mode.clear()
        self._relim_requested = True

    @property
    def data(self):
        return self.viewer_state.reference_data

    @property
    def profile_axis(self):
        # XXX make this settable
        # defaults to the non-xy axis with the most channels
        try:
            slc = self.viewer_state.wcsaxes_slice[::-1]
        except AttributeError:
            return None
        candidates = [i for i, s in enumerate(slc) if s not in ['x', 'y']]
        return max(candidates, key=lambda i: self.data.shape[i])

    def _recenter_grips(self):
        for ctx in self._contexts:
            ctx.recenter(self.axes.get_xlim())

    def _extract_subset_profile(self, subset):
        slc = self.viewer_state.slices
        try:
            x, y = Extractor.subset_spectrum(subset,
                                             self.viewer_state.display_attribute,
                                             slc,
                                             self.profile_axis)
        except IncompatibleAttribute:
            return

        self._set_profile(x, y)

    def _update_from_roi(self, roi):
        data = self.data
        att = self.viewer_state.layers[0].attribute
        slc = self.viewer_state.wcsaxes_slice[::-1]

        if data is None or att is None:
            return

        zax = self.profile_axis

        x, y = Extractor.spectrum(data, att, roi, slc, zax)
        self._set_profile(x, y)

    def _update_profile(self, *args):
        roi = self.mouse_mode.roi()
        return self._update_from_roi(roi)

    def _move_profile(self, *args):
        if self.mouse_mode._roi_tool._scrubbing:
            self._update_profile(args)

    def _set_profile(self, x, y):
        data = self.data

        xid = data.get_world_component_id(self.profile_axis)
        units = data.get_component(xid).units
        xlabel = str(xid) if units is None else '%s [%s]' % (xid, units)

        xlim = self.axes.get_xlim()
        self.profile.set_xlabel(xlabel)
        self.profile.set_profile(x, y, color='k')

        # relim x range if requested
        if self._relim_requested:
            self._relim_requested = False
            self.axes.set_xlim(np.nanmin(x), np.nanmax(x))

        # relim y range to data within the view window
        self.profile.autoscale_ylim()

        if self.axes.get_xlim() != xlim:
            self._recenter_grips()

        self.axes.figure.canvas.draw()
        self.show()

    def _move_below_image_viewer(self):
        rect = self.image_viewer.frameGeometry()
        pos = rect.bottomLeft()
        self._mdi_wrapper.setGeometry(pos.x(), pos.y(),
                                      rect.width(), 300)

    def show(self):
        if self.widget.isVisible():
            return
        self._move_below_image_viewer()
        self.widget.show()

    def hide(self):
        if hasattr(self, '_mdi_wrapper'):
            self._mdi_wrapper.close()
        else:
            self.widget.close()

    def _get_modes(self, axes):
        return [self.mouse_mode]



from __future__ import absolute_import, division, print_function

from mock import MagicMock

from qtpy.QtCore import Qt
from qtpy import PYQT5
from glue.core import Data, Hub
from glue.core.layer_artist import LayerArtistBase as _LayerArtist

from ..layer_artist_model import LayerArtistModel, LayerArtistView


class LayerArtist(_LayerArtist):

    def update(self, view=None):
        pass

    def clear(self):
        pass

    def redraw(self):
        pass


def setup_model(num):
    mgrs = [LayerArtist(Data(label=str(i))) for i in range(num)]
    model = LayerArtistModel(mgrs)
    return model, mgrs


def test_row_count():
    for n in range(4):
        assert setup_model(n)[0].rowCount() == n


def test_row_label():
    model, mgrs = setup_model(5)
    for i in range(5):
        assert model.row_label(i) == mgrs[i].layer.label


def test_add_artist_updates_row_count():
    mgrs = [LayerArtist(Data(label='A'))]
    model = LayerArtistModel(mgrs)
    model.add_artist(0, LayerArtist(Data(label='B')))
    assert model.rowCount() == 2


def test_add_artist_updates_artist_list():
    mgrs = [LayerArtist(Data(label='A'))]
    model = LayerArtistModel(mgrs)
    model.add_artist(0, LayerArtist(Data(label='B')))
    assert len(mgrs) == 2


def test_valid_remove():
    mgr = MagicMock(spec_set=LayerArtist)
    mgrs = [mgr]
    model = LayerArtistModel(mgrs)
    remove = model.removeRow(0)
    assert remove
    assert mgr not in mgrs


def test_invalid_remove():
    mgr = MagicMock(spec_set=LayerArtist)
    mgrs = [mgr]
    model = LayerArtistModel(mgrs)
    remove = model.removeRow(1)
    assert not remove
    assert mgr in mgrs


def test_artist_cleared_on_remove():
    mgr = MagicMock(spec_set=LayerArtist)
    mgrs = [mgr]
    model = LayerArtistModel(mgrs)

    model.removeRow(0)
    mgr.clear.assert_called_once_with()


def test_change_label():
    model, (mgr,) = setup_model(1)

    lbl = mgr.layer.label
    model.change_label(0, 'new label')
    assert mgr.layer.label != lbl


def test_change_label_invalid_row():
    model, (mgr,) = setup_model(1)

    lbl = mgr.layer.label
    model.change_label(1, 'new label')
    assert mgr.layer.label == lbl


def test_flags():
    model, layer_artists = setup_model(1)

    expected = (Qt.ItemIsEditable |
                Qt.ItemIsDragEnabled |
                Qt.ItemIsEnabled |
                Qt.ItemIsSelectable |
                Qt.ItemIsUserCheckable)

    if PYQT5:
        expected |= Qt.ItemNeverHasChildren

    assert model.flags(model.index(0)) == expected


def test_move_artist_empty():
    mgrs = []
    model = LayerArtistModel(mgrs)
    model.move_artist(None, 0)

    assert mgrs == []


def test_move_artist_single():
    m0 = LayerArtist(Data(label="test 0"))
    mgrs = [m0]

    model = LayerArtistModel(mgrs)
    model.move_artist(m0, 0)
    assert mgrs == [m0]
    model.move_artist(m0, -1)
    assert mgrs == [m0]
    model.move_artist(m0, 1)
    assert mgrs == [m0]
    model.move_artist(m0, 2)
    assert mgrs == [m0]


def test_move_artist_two():
    model, mgrs = setup_model(2)
    m0, m1 = mgrs

    model.move_artist(m0, 0)
    assert mgrs == [m0, m1]

    model.move_artist(m0, 1)
    assert mgrs == [m0, m1]

    model.move_artist(m0, 2)
    assert mgrs == [m1, m0]

    model.move_artist(m0, 0)
    assert mgrs == [m0, m1]


def test_move_artist_three():
    model, mgrs = setup_model(3)
    m0, m1, m2 = mgrs

    model.move_artist(m0, 0)
    assert mgrs == [m0, m1, m2]

    model.move_artist(m0, 1)
    assert mgrs == [m0, m1, m2]

    model.move_artist(m0, 2)
    assert mgrs == [m1, m0, m2]
    model.move_artist(m0, 0)

    model.move_artist(m0, 3)
    assert mgrs == [m1, m2, m0]
    model.move_artist(m0, 0)

    model.move_artist(m2, 0)
    assert mgrs == [m2, m0, m1]


def test_move_updates_zorder():
    m0 = LayerArtist(Data(label='test 0'))
    m1 = LayerArtist(Data(label='test 1'))
    m2 = LayerArtist(Data(label='test 2'))
    m0.zorder = 10
    m1.zorder = 20
    m2.zorder = 30

    mgrs = [m0, m1, m2]
    model = LayerArtistModel(mgrs)

    model.move_artist(m2, 0)
    assert m2.zorder == 30
    assert m0.zorder == 20
    assert m1.zorder == 10


def test_check_syncs_to_visible():
    m0 = LayerArtist(Data(label='test 0'))
    m0.artists = [MagicMock()]
    mgrs = [m0]
    model = LayerArtistModel(mgrs)

    m0.visible = True
    assert m0.visible
    assert model.data(model.index(0), Qt.CheckStateRole) == Qt.Checked
    m0.visible = False
    assert not m0.visible
    assert model.data(model.index(0), Qt.CheckStateRole) == Qt.Unchecked

    model.setData(model.index(0), Qt.Checked, Qt.CheckStateRole)
    assert m0.visible


def test_data():

    model, mgrs = setup_model(3)
    idx = model.index(3)
    assert model.data(idx, Qt.DisplayRole) is None
    idx = model.index(1)
    assert model.data(idx, Qt.DisplayRole) == model.row_label(1)
    assert model.data(idx, Qt.EditRole) == model.row_label(1)


class TestLayerArtistView(object):

    def setup_method(self, method):
        self.model, self.artists = setup_model(2)
        self.hub = Hub()
        self.view = LayerArtistView(hub=self.hub)
        self.view.setModel(self.model)

    def test_current_row(self):
        for row in [0, 1]:
            idx = self.model.index(row)
            self.view.setCurrentIndex(idx)
            self.view.current_row() == row
            assert self.view.current_artist() is self.artists[row]



#!/usr/bin/env python
"""
Script used to create template config.py files for Glue
"""

from __future__ import absolute_import, division, print_function

import os
import sys
from shutil import copyfile

import glue
from glue.external.six.moves import input


def get_clobber():
    result = None
    result = input("\nDestination file exists. Overwrite? [y/n] ")
    while result not in ['y', 'n']:
        print("\tPlease choose one of [y/n]")
        result = input("\nDestination file exists. Overwrite? [y/n] ")

    return result == 'y'


def main():

    # Import at runtime because some tests change this value. We also don't
    # just import the function directly otherwise it is cached.
    from glue import config
    dest = config.CFG_DIR

    if not os.path.exists(dest):
        print("Creating directory %s" % dest)
        os.makedirs(dest)

    infile = os.path.join(glue.__path__[0], 'default_config.py')
    outfile = os.path.join(dest, 'config.py')

    print("Creating file %s" % outfile)

    if os.path.exists(outfile):
        clobber = get_clobber()
        if not clobber:
            print("Exiting")
            sys.exit(1)

    copyfile(infile, outfile)

if __name__ == "__main__":
    main()



# pylint: disable=I0011,W0613,W0201,W0212,E1101,E1103

from __future__ import absolute_import, division, print_function

import warnings

from glue.config import viewer_tool
from glue.viewers.common.qt.data_viewer import DataViewer
from glue.viewers.common.qt.tool import Tool
from glue.viewers.common.qt.toolbar import BasicToolbar
from glue.core.tests.util import simple_session


@viewer_tool
class ExampleTool1(Tool):

    tool_id = 'TEST1'
    tool_tip = 'tes1'
    icon = 'glue_square'
    shortcut = 'A'


@viewer_tool
class ExampleTool2(Tool):

    tool_id = 'TEST2'
    tool_tip = 'tes2'
    icon = 'glue_square'
    shortcut = 'A'


class ExampleViewer2(DataViewer):

    _toolbar_cls = BasicToolbar
    tools = ['TEST1', 'TEST2']

    def __init__(self, session, parent=None):
        super(ExampleViewer2, self).__init__(session, parent=parent)


def test_duplicate_shortcut():
    session = simple_session()
    with warnings.catch_warnings(record=True) as w:
        ExampleViewer2(session)
    assert len(w) == 1
    assert str(w[0].message) == ("Tools 'TEST1' and 'TEST2' have the same "
                                 "shortcut ('A'). Ignoring shortcut for 'TEST2'")
